local MAJOR = "StarLibCore-1.0" 
local MINOR = 1

local LibCore = LibStub:NewLibrary(MAJOR, MINOR)
if not LibCore then return end
local LibTimer = LibStub("StarLibTimer-1.0", true)
assert(LibTimer, MAJOR .. " requires StarLibTimer-1.0")
local LibError = LibStub("StarLibError-1.0", true)
assert(LibError, MAJOR .. " requires StarLibError-1.0")
local LibCFG = LibStub("StarLibCFG-1.0", true)
assert(LibCFG, MAJOR .. " requires StarLibCFG-1.0")
local WidgetText = LibStub("StarLibWidgetText-1.0", true)
assert(WidgetText, MAJOR .. " requires StarLibWidgetText-1.0")
local WidgetBar = LibStub("StarLibWidgetBar-1.0", true)
--assert(WidgetBar, MAJOR .. " requires StarLibWidgetBar-1.0")
local WidgetIcon = LibStub("StarLibWidgetIcon-1.0", true)
--assert(WidgetIcon, MAJOR .. " requires StarLibWidgetIcon-1.0")
local WidgetHistogram = LibStub("StarLibWidgetHistogram-1.0", true)
--assert(WidgetHistogram, MAJOR .. " requires StarLibWidgetHistogram-1.0")
local WidgetBignums = LibStub("StarLibWidgetBignums-1.0", true)
--assert(WidgetBignums, MAJOR .. " requires StarLibWidgetBignums-1.0")
local WidgetKey = LibStub("StarLibWidgetKey-1.0", true)
--assert(WidgetKey, MAJOR .. " requires StarLibWidgetKey-1.0")
local WidgetTimer = LibStub("StarLibWidgetTimer-1.0", true)
--assert(WidgetTimer, MAJOR .. " requires StarLibWidgetTimer-1.0")
local WidgetScript = LibStub("StarLibWidgetScript-1.0", true)
--assert(WidgetScript, MAJOR .. " requires StarLibWidgetScript-1.0")
local PluginRangeCheck = LibStub("StarLibPluginRangeCheck-1.0")
assert(PluginRangeCheck, MAJOR .. " requires StarLibRangeCheck-1.0")

local pool = setmetatable({}, {__mode = "k"})
local objects = {}
local objectDicts = {}

if not LibCore.__index then
	LibCore.__index = LibCore
end

local TRANSITION_RIGHT = 0
local TRANSITION_LEFT = 1
local TRANSITION_BOTH = 2
local TRANSITION_UP = 3
local TRANSITION_DOWN = 4
local TRANSITION_TENTACLE = 5
local TRANSITION_ALPHABLEND = 6

-- @name StarLibCore:New
-- @usage StarLibCore:New(visitor, environment, name, config, typeOf, lcd, errorLevel)
-- @parma visitor App table
-- @return A new StarLibCore object
function LibCore:New(visitor, environment, name, config, typeOf, lcd, errorLevel)
	
	local obj = next(pool)

	if obj then
		pool[obj] = nil
	else
		obj = {}
	end

	setmetatable(obj, self)
	
	tinsert(objects, obj)
	
	obj.visitor = visitor
	obj.environment = environment
	obj.typeOf = typeOf
	obj.name = name
	obj.layout_timeout = 10000
	obj.timeout = 500
	obj.is_transitioning = true
	obj.timer = LibTimer:New("LibCore.timer", obj.timeout)
	obj.transition_timer = LibTimer:New("LibCore.transition_timer", obj.layout_timeout)
	obj.CFG = LibCFG:New(obj, config, environment)
	obj.lcd = lcd
	obj.error = LibError:New(errorLevel)
	obj.RangeCheck = PluginRangeCheck:New(environment)
		
	return obj	
	
end

-- @name StarLibCore:Del
-- @usage StarLibCore:Del([core]) or object:Del()
-- @param core An optional core object
-- @return Nothing
function LibCore:Del(core)
	if core and type(core) == "table" then
		pool[core] = true
		for i, v in ipairs(objects) do
			if v == core then
				tremove(objects, i)
			end
		end
	else
		for i, v in ipairs(objects) do
			if v == self then
				tremove(objects, i)
			end
		end
		pool[self] = true
	end
end

function LibCore:ResizeText(str, cols)
	if strlen(str) < cols then
		for i = strlen(str) + 1, cols do
			str = str .. " "
		end
	end
	return str:sub(1, cols)
end

function LibCore:ReplaceText(text1, pos1, text2, pos2)
	local char = text2:sub(pos2, pos2)
	
	local left = text1:sub(1, pos1 - 1)
	local right = text1:sub(pos1 + 1)
	
	return left .. char .. right
end

function LibCore:Memcopy(to, from, size)
	for i = 0, size - 1 do
		to = self:ReplaceText(to, i, from, i)
	end
	return to
end

function LibCore:CFGSetup()
	local i = 1
	local str = ""
	

	--app_->ProcessVariables(CFG_Get_Root(), (Evaluator *)this);

	local section = CFG_Fetch_Raw(self.CFG:GetRoot(), self.name);
	if( not section) then
		LCDError(("Display <%s> doesn't exist in configuration."):format( self.name))
		return
	end

	self.layout_timeout = self.CFG:Fetch(section, "layout-timeout", 10000);

	self.transition_speed = self.CFG:Fetch(section, "transition-speed", 200);

	local layout = self.CFG:FetchRaw(section, "layout0");

	if not layout then
		layout = self.CFG:FetchRaw(section, "layout")
		if not layout then
			LCDError("Couldn't find a layout.")
			return
		end
	end
	
	local i = 1
	while(layout) do
		tinsert(self.layouts, layout)
		local str = "layout" .. i
		layout = self.CFG:FetchRaw(section, str);
		i = i + 1
	end

	if not layout then
		layout = self.CFG:FetchRaw(section, "layout")
		if not layout then
			LCDError("Couldn't find a layout.")
			return
		end
	end

	local widget = self.CFG:FetchRaw(section, "widget0");

	i = 1;
	while(widget and type(widget) == "string") do
		tinsert(self.static_widgets, widget)
		local str = "widget" + i
		widget = self.CFG:FetchRaw(section, str);
		i = i + 1
	end

	i = 0
	while i < #self.layouts do
		layout = self.CFG:FetchRaw(self.CFG.root, self.layouts[i]);
		if not layout then
			LCDError(("Missing layout <%s>"):format( layouts[i]));
		end
		if(layout) then 
			--Json::Value *val = CFG_Fetch(layout, "keyless", new Json::Value(0));
			--if(val->asInt()) {
			--	keyless_layouts_[layouts_[i]] = true;
			--}

			local layer = 0
			while(layer < self.lcd.LAYERS) do
				local str = "layer" .. layer + 1
				local cfg_layer = self.CFG:FetchRaw(layout, str);
				if(cfg_layer) then
					local row = 0
					while row < self.lcd.LROWS do
						str = "row" .. row + 1
						local cfg_row = self.CFG:FetchRaw(cfg_layer, str)
						if cfg_row then
							local col = 0
							while(col < self.lcd.LCOLS) do
								str = "col" .. col + 1
								local cfg_col = self.CFG:FetchRaw(cfg_row, str);

								if(cfg_col and type(cfg_col) == "string") then
									local w = {}
									w.key = cfg_col
									w.row = row
									w.col = col
									w.layer = layer;
									tinsert(self.widget_templates[self.layouts[i]], w)
								end
								col = col + 1
							end
						end
						row = row + 1
					end
				end
				layer = layer + 1
			end

			local row = 0
			while row < self.lcd.LROWS do
				local str = "row" .. row + 1
				local cfg_row = self.CFG:FetchRaw(layout, str);
				if(cfg_row) then
					local col = 0
					while col < self.lcd.LCOLS do
						local str = "col" .. col + 1
						cfg_col = self.CFG:FetchRaw(cfg_row, "col" .. col + 1);

						if(cfg_col and type(cfg_col) == "string") then
							local w = {}
							w.key = cfg_col
							w.row = row;
							w.col = col;
							w.layer = 0;
							tinsert(self.widget_templates[self.layouts[i]], w)
						end
						col = col + 1
					end
				end
				row = row + 1
			end
		end
		i = i + 1
	end

	local j = 0
	while j < #self.static_widgets do
		local w = {}
		w.key = self.static_widgets[j];
		w.row = 0;
		w.col = 0;
		tinsert(self.widget_templates[self.name], w)
		j = j + 1
	end

end

function LibCore:BuildLayouts()
	local name
	for k, v in pairs(self.widget_templates) do
		local widgets = v
		for i = 0, #widgets do
			local widget_v = self.CFG:FetchRaw(self.CFG.root, widgest[i].key)
			if(not widget_v) then
				self.error(("No widget named <%s>"):format(widgets[i].key.c_str()))
			else
				local typeOf = self.CFG:Fetch_Raw(widget_v, "typeOf");
				if(not typeOf) then
					self.error(("Widget <%s> has no typeOf!"):format( widgets[i].key.c_str()))
				else
					local widget
	 
					local name;
					local j = 0;
					name = k .. ":" .. widgets[i].key .. ":" .. j
	 --[[
					for k, v in pairs(self.widgets) do
					while(self.widgets.find(name) ~= self.widgets.end()) do
						if v.name == name then
							j = j + 1
							name = k .. ":" .. widgets[i].key .. ":" .. j
						end
					end
	 ]]
					if(typeOf == "text") then
						widget = WidgetText:New(self, name, widget_v, widgets[i].row, widgets[i].col, widgets[i].layer)
					elseif (typeOf == "bar") then
						--widget = WidgetBar:New(self, name, widget_v, widgets[i].row, widgets[i].col, widgets[i].layer)
					elseif (typeOf == "icon") then
						--widget = WidgetIcon:New(self, name, widget_v, widgets[i].row, widgets[i].col, widgets[i].layer)
					elseif (typeOf == "histogram") then
						--widget = WidgetHistogram:New(self, name, widget_v, widgets[i].row, widgets[i].col, widgets[i].layer)
					elseif (typeOf == "bignums") then
						--widget = WidgetBignums:New(self, name, widget_v, widgets[i].row, widgets[i].col, widgets[i].layer)
					elseif (typeOf == "key") then
						--widget = WidgetKey:New(self, name, widget_v)
					elseif (typeOf == "timer") then
						--widget = WidgetTimer:New(self, name, widget_v) 
					elseif (typeOf == "script") then
						--widget = WidgetScript:New(self, name, widget_v)
					else
						 self.error("Unknown widget typeOf: %s", type(typeOf))
					end
					if(widget) then
						self.widgets[name] = widget;
					end
				end
			end
		end
	end
end

function LibCore:StartLayout(key)
    if(key == "") then
		self.gen = newGenerator(self.current_layout, self.last_layout_)
    else
        self.current_layout = key;
	end

    self.error("StartLayout: %s", current_layout_.c_str());
	self.lcd:TextClear()
	
    --emit static_cast<LCDEvents *>(wrapper_)->_LayoutChangeBefore();
	
	local widgets = self.widgets
	for k, w in pairs(widgets) do

        if(not(self.current_layout_ ~= w.layout_base and w.name ~= self.name) ) then

			if self.type.widget_special and Utils.band(w.type, WIDGET_TYPE_SPECIAL) then
				w:SetupChars();
			end
			w:Start();
		end
    end

	for k, w in pairs(widgets) do
		if self.name == w.layout_base then
			w:Start()
		end
    end

    --emit static_cast<LCDEvents *>(wrapper_)->_LayoutChangeAfter();

	local timeout = self.visitor.CFG:Fetch(self.visitor, self.visitor.CFG.root, self.current_layout .. ".timeout", self.layout_timeout_);

    if(timeout > 0) then
        self.timer:Start(timeout);
	end


    self.is_transitioning = false;
end

function LibCore:StopLayout(layout) 
	local widgets = self.widgets
    for k, w in pairs(widgets) do
        if(layout == w.layout_base ) then
			w:Stop();
		end
    end
end

function LibCore:ChangeLayout() 
    LCDError("ChangeLayout");
	local t = self.CFG:FetchRaw(self.CFG.root, self.current_layout .. ".transition")
    if(not t) then
        self:StopLayout(self.current_layout);
        self:StartLayout();
    else 
        self:StartTransition(t);
	end
    
end

function LibCore:Transition(i) 
    if(self.is_transitioning) then
        return;
	end
    self.gen = layoutGenerator(i, this);
    self:ChangeLayout();
    LCDError("LibCore::Transition " .. i);
end

function LibCore:StartTransition(transition)
    if(not transition) then
        LCDError(("Transition not defined <%s>"):format( current_layout_.c_str()));
        self:StartLayout();
        return;
    end
	local t
	local case = transition.sub(1, 1)
	if case == "R" then
		t = TRANSITION_RIGHT
	elseif case == "L" then
		t = TRANSITION_LEFT
	elseif case == "B" then
		t = TRANSITION_BOTH
	elseif case == "U" then
		t = TRANSITION_UP
	elseif case == "D" then
		t = TRANSITION_DOWN
	else
		t = TRANSITION_RIGHT
	end
    self.is_transitioning = true;
    self.gen:GetNext(current_layout_, last_layout_);
    --lcd_->SignalTransitionStart(current_layout_);
    self:StartLayout(self.current_layout);
    self.direction = t;
	self.transition_timer:Start(self.transition_speed)
    self:LayoutTransition();
    LCDError(("Transition started -- speed: %d"):format(self.transition_speed));
end

function LibCore:LayoutTransition()
	self.lcd:Transition()
end

function LibCore:TransitionFinished()
    self.transition_time:Stop();
    self:StopLayout(self.last_layout);
    self.is_transitioning = false;
    --//lcd_->SignalTransitionEnd();
end

--[[std::string LibCore::CFG_Key() {
    return key_;
}]]

--[[
function LibCore::KeypadEvent(key) 
    LCDError(("KeypadEvent(%d) %s"):format(key, current_layout_.c_str()));

    std::map<std::string, bool>::iterator it = 
        keyless_layouts_.find(current_layout_);

    if(it != keyless_layouts_.end())
        return;

    for(std::map<std::string, Widget *>::iterator w = widgets_.begin();
        w != widgets_.end(); w++) {
        if( w->second->GetType() & WIDGET_TYPE_KEYPAD )
            ((WidgetKey*)w->second)->KeyPressed(key);
    }
}
]]

function LibCore:ResizeLCD(rows, cols)
    self:StopLayout(self.current_layout);
    local old_rows = lcd.LROWS;
    local old_cols = lcd.LCOLS;
    if(self:ResizeLCD(rows, cols) == 0) then
        --emit static_cast<LCDEvents *>(wrapper_)->_ResizeLCD(rows, cols, old_rows, old_cols);
    else
        LCDError("LibCore::ResizeLCD: Unable to resize LCD");
        return -1;
    end
    self:StartLayout(self.current_layout);
    return 0;
end

function LibCore:SelectLayout(layout) 
    if(self.is_transitioning) then
        return;
	end
	local i = 0
	while i < #self.layouts do
		if self.layouts[i] == layout then
			self:StopLayout(self.current_layout)
			self:StartLayout(layout)
		end
	end
end

function LibCore:RemoveWidget(name) 
    if(type(self.widgets[name]) ~= "nil") then
		self.widgets[name]:Del()
		self.widgets:remove(self.widgets[name])
        return 0;
    end
    return -1;
end

function LibCore:AddWidget(layout, row, col, layer, object) 
	local name = layout .. ":" .. "script_widget"
	
    local i = 0;
    while(widgets[name]) do
		name = layout .. ":" .. "script_widget" .. ":" .. i
		i = i + 1
    end
    
    local type = self.CFG:FetchRaw(object, "type");

    if(type) then
           local widget
           if(type == "text") then
               widget = WidgetText:New(self, name, root, row, col, layer);
           elseif (type == "bar") then
               widget = WidgetBar:New(self, name, root,row, col, layer);
           elseif (type == "icon") then
               widget = WidgetIcon:New(self, name, root, row, col, layer);
           elseif (type == "histogram") then
               widget = WidgetHistogram:New(self, name, root, row, col, layer);
           elseif (type == "bignums") then
               widget = WidgetBignums:New(self, name, root, row, col, layer);
           elseif (type == "key") then
               widget = WidgetKey:New(self, name, root);
           elseif (type == "timer") then
               widget = WidgetTimer:New(self, name, root);
           elseif (type == "script") then
               widget = WidgetScript:New(self, name, root);
           else
               LCDError("Unknown widget type: %s", type);
           end
           if(widget) then
               self.widgets[name] = widget;
               widget:Start();
           end
           return name;
    else
        LCDError(("Widget has no type <%s>"):format(object.c_str()));
    end
    return "";
end

--[[
int LibCore::MoveWidget(std::string widget, int row, int col) {
    return 0;
}
]]
